import threading
import logging
import operator

import locked_queue
import queen_dir

import queen
import worker

# imports for high-level scheduling
import random
import job_list



# configures debug logging?
logging.basicConfig(filename='/tmp/hive.log', level=logging.DEBUG,
    format='%(asctime)s (%(threadName)-2s) %(message)s')

class Hive(object):

	# creates the 'global' queen directory using QueenDir
	def __init__(self):
		self.queens=queen_dir.QueenDir()

	# creates a new queen object
	def create_queen(self, name, default_worker=worker.Worker):
		logging.debug("Creating a queen with name \'%s\'..." %name)
		lock = locked_queue.LockedQueue()
		sched = job_list.JobList()
		cq = queen.Queen(name, lock, sched, default_worker)
		c = threading.Thread(name="%s:queen"%name,target=cq.main)
		self.queens.add_queen(name, c, lock, sched)

	# kill gracefully, sending a special "die"-marked queue
	def kill_queen(self,name):
		# code will
		if self.queens.get_queen(name).isAlive():
			if self.queens.get_stat(name) != "ded":
				logging.debug("Waiting for the lock in order to kill %s..."%name)
				self.queens.get_lock(name).acquire()
				logging.debug("Sending death command to %s..."%name)
				self.queens.get_lock(name).append('die')
				self.queens.set_stat(name, "ded")
				self.queens.get_lock(name).cond.notify()
				self.queens.get_lock(name).release()
			else:
				logging.debug("Kill command for queen \'%s\' failed: already marked for death." %name)
		else:
			logging.debug("ERROR: I tried to kill a queen (%s) who was unstarted!"%name)

	# starts the queen using the QueenDir, but only if the queen is
	# not alive, else prints an error
	def start_queen(self, name):
		if not self.queens.get_queen(name).isAlive():
			logging.debug("Starting the queen with name \'%s\'..."%name)
			self.queens.get_queen(name).start()
		else:
			logging.debug("ERROR: I tried to start a queen (%s) who was already alive!"%name)

	# Appends some instructions to the locked queue as a tuple
	# of the instruction value(s) and the worker class to use.
	# The format of the instructions will differ depending
	# on the particular implementation of the queen.
	# All instructions provided will use the worker type specified.
	# Accepts lists gracefully as individual commands (probably).
	def instruct_queen(self, name, instructions, worker_type='default'):
		self.queens.get_lock(name).acquire()
		if type(instructions) == list:
			for i in instructions:
				self.queens.get_lock(name).append((i,worker_type))
		else:
			self.queens.get_lock(name).append((instructions,worker_type))
		self.queens.get_lock(name).cond.notify()
		self.queens.get_lock(name).release()

	# die gracefully, returns a dict of results also
	def die(self):
		logging.debug("Trying to die...")
		for queen in self.queens.enum_queens():
			self.kill_queen(queen)
		for queen in self.queens.enum_queens():
			self.queens.get_queen(queen).join()
			if not self.queens.get_queen(queen).isAlive():
				logging.debug("Joined: %s" % queen)
			else:
				assert not self.queens.get_queen(queen).isAlive(), "Joined the queen \'%s\', but the thread is still alive." % queen
		results = {}
		for queen in self.queens.enum_queens():
			results[queen] = self.get_result(queen)
		logging.debug("RIP everyone")
		return results

	# log results from a queen
	def get_result(self, name):
		self.queens.set_result(name, self.queens.get_lock(name).access())
		return self.queens.get_result(name)

	# return list of queens?
	def get_queens(self):
		return self.queens.enum_queens()






""" QueenDir (Queen Directory)

This module contains a nested dictionary data structure that is being
used to store objects representing threaded queue handlers.
"""

class QueenDir(object):
    """ Queen Directory

    This class is used as an interface for storing and managing
    objects of interest and their related attributes and components.
    An example present here is the storage of a queen who manages
    launching individual worker objects and is on a separate thread
    from MainThread. As such, the task of managing the queens is much
    simpler when the object representing the thread is able to be
    recalled by name, along with its related locks and other information.
    """

    def __init__(self):
        """ Initialization

        Attributes:
            lib [dict]: A simple hashmap used to store the queens
                alongside their other details
        """
        self.lib = {}

    def add_queen(self, name, queen, lock, sched):
        """ AddQueen (Adding a queen entry)

        This method is used to create a new entry for queen while adding
        the appropriate attributes to the dictionary under that index.
        A nested dictionary structure results, as the entries are
        organized into dictionaries.

        Args:
            name (string): A human-readable value to represent the entry
            queen (object): An instantiation of a threaded class
            lock (object): An instantiation of a threading lock
            sched (object): An instantiation of a task scheduler

        Attributes:
            lib[name]['stat'] (string): A value to represent other
                properties of interest, such as being gravebound
                (having received a kill command)
            lib[name]['result'] (object): Some object which the
                queen stores, typically containing useful information
                about the results of work done by her workers
        """
        self.lib[name] = {'queen':queen, 'lock':lock,
            'sched':sched, 'stat':None, 'result':None}

    def get_lock(self, name):
        """ getLock

        Returns the object indexed under lib[name]['lock']

        Args:
            name (string): The value representing the entry of interest

        Returns:
            lib[name]['lock'] (object): The object mapped to 'lock' for
                a given queen name
        """
        return self.lib[name]['lock']

    def get_queen(self, name):
        """ getQueen

        Returns the queen object indexed under lib[name]['queen'].

        Args:
            name (string): The value representing the entry of interest

        Returns:
            lib[name]['queen'] (object): The queen object belonging to
                a given queen name
        """
        return self.lib[name]['queen']

    def enum_queens(self):
        """ enumQueens (Enumerate queens)

        Returns a list of names of entries. Particularly useful for
        accessing all queens when performing operations like programmed
        controller death.

        Returns:
            lib.keys() (list of strings): A list that includes the name
                for each queen entry
        """
        return self.lib.keys()

    def set_stat(self, name, string):
        """ setStat

        Stores a string under the index lib[name]['stat'].

        Args:
            name (string): The value representing the entry of interest
            string (string): A string to be stored as the queen's status
        """
        self.lib[name]['stat'] = string

    def get_stat(self, name):
        """ getStat

        Returns the value for 'stat'.

        Args:
            name (string): The value representing the entry of interest

        Returns:
            lib[name]['stat'] (string): The status string belonging to
                a given queen name
        """
        return self.lib[name]['stat']

    def get_sched(self, name):
        """ getSched

        Returns the schedue object indexed under lib[name]['sched'].
        This method is currently unused, but is intended for use
        as part of an update to add transparency to job execution.

        Args:
            name (string): The value representing the entry of interest

        Returns:
            lib[name]['sched'] (object): The schedule object belonging to
                a given queen name
        """
        return self.lib[name]['sched']

    def get_result(self, name):
        """ getResult

        Returns the result object indexed under lib[name]['result'].

        Args:
            name (string): The value representing the entry of interest

        Returns:
            lib[name]['result'] (object): The queen object belonging to
                a given queen name
        """
        return self.lib[name]['result']

    def set_result(self, name, val):
        """ setResult

        Stores a result object under the index lib[name]['result'].

        Args:
            name (string): The value representing the entry of interest
            val (object): Some value to be stored as the result to be
                recalled later
        """
        self.lib[name]['result'] = val



import locked_queue
import logging
import time
import threading

# configures debug logging?
logging.basicConfig(filename='/tmp/queen.log', level=logging.DEBUG,
    format='%(asctime)s (%(threadName)-2s) %(message)s',)


class Queen(object):

	# sets its lock, schedule, name, and default worker type
	def __init__(self, name, lock, sched, default_worker):
		self.lock = lock
		self.sched = sched
		self.name = name
		self.default_worker = default_worker

	# wait for the locked queue to be unlocked, and when it is unlocked
	# append the queue contents to a working job queue
	def main(self):
		f = int(1)
		logging.debug('Starting the queen...')
		local_lock = locked_queue.LockedQueue()
		self.lock.acquire()
		logging.debug('Acquired the lock for the first time.')
		workers=[]
		die = False
		while True:
			q = self.lock.enum()
			for i in q:
				if i == 'die':
					die = True
				else:
					# pick instruction value from tuple
					instruction = i[0]
					
					# look at worker variable, if it specifies default
					# use the default passed on queen creation, else
					# use the class passed to it
					if i[1] == 'default':
						worker_var = self.default_worker
					else:
						worker_var = i[1]

					# build and start the worker
					cw = worker_var(local_lock,instruction)
					c = threading.Thread(name="%s:worker%d"%(self.name,f),target=cw.run)
					f += 1
					c.start()
					# janky way of avoiding threads starting out of order
					time.sleep(0.1)
					workers.append(c)
			if die == True:
				logging.debug('Death command received.')
				break
			else:
				self.lock.clear()
				self.lock.cond.notify()
				logging.debug('Apiary notified. Releasing...')
				self.lock.cond.wait()
				logging.debug('Reacquired the lock.')

		# dies gracefully, joining all worker threads
		logging.debug("Waiting for remaining workers...")
		for i in workers:
			i.join()
		logging.debug("No more workers in line.")

		# clears the locked queue and then appends worker results to it
		logging.debug("Sending results back to the hive...")
		self.lock.clear()
		self.lock.append(local_lock.enum())
		self.lock.release()
		# dies
		logging.debug("Queen died (gracefully, one hopes).")




import threading
import time
import sys

from busybees import worker
from busybees import hive

import pash

class ErrWorker(worker.Worker):
    def work(self, command):
        proc = pash.ShellProc()
        proc.run(command)
        return "Exit code: %s" % proc.get_val('exit_code')

def test_hive():
    apiary = hive.Hive()
    apiary.create_queen('A1')
    apiary.create_queen('A2')
    apiary.start_queen('A1')
    apiary.start_queen('A2')

    job1 = ["iscsiadm -m discovery -t st -p 192.168.88.110",
            "iscsiadm -m discovery -t st -p 192.168.90.110",
            "iscsiadm -m discovery -t st -p 192.168.88.110"]
    apiary.instruct_queen('A1', job1, ErrWorker)

    job2 = ["ls -l ~", "date", "cal"]
    apiary.instruct_queen('A2', job2)

    apiary.kill_queen('A1')

    time.sleep(3)

    results = apiary.die()

    for key in results.keys():
        for i in results[key]:
            assert i != '' and i != None



import threading

# a data structure with a queue, lock, and condition


class LockedQueue(object):

	def __init__(self):
		self.queue = []
		self.lock = threading.Lock()
		self.cond = threading.Condition(self.lock)

	# explicit lock acquisition
	def acquire(self):
		self.lock.acquire()

	# explicit lock releasing
	def release(self):
		self.lock.release()

	# acquires the lock then returns the queue
	def access(self):
		self.lock.acquire()
		return self.queue

	# returns the queue
	def enum(self):
		return self.queue

	# append to the queue, additional list filter logic
	def append(self, obj):
		if type(obj) == list:
			self.queue.extend(obj)
		else:
			self.queue.append(obj)

	# clears the queue, used for tentative jobs, etc.
	def clear(self):
		self.queue = []





# each entry contains three pieces of information:
#	1) Job id number, which must be managed and remembered
#	   by the calling function, preferably via RNG
#	2) Command that should be run, in bash syntax
#	3) The status of the job, where: (integers deprecated, do not use)
#		0 = scheduled
#		1 = current
#		2 = completed
#		3 = cancelled

class JobList(object):

	def __init__(self):
		self.queue = []

	def add(self, idnum, command):
		entry = [idnum, command, 'scheduled']
		self.queue.append(entry)

	def list_all(self):
		return self.queue

	def cancel(self, idnum):
		index = self.find_tuple(idnum, self.queue)
		assert self.queue[index][2] == 'scheduled', "Job looks like it wasn't waiting."
		self.queue[index][2] = 'cancelled'

	def curr(self, idnum):
		index = self.find_tuple(idnum, self.queue)
		assert self.queue[index][2] == 'scheduled', "Job looks like it wasn't waiting."
		self.queue[index][2] = 'current'

	def done(self, idnum):
		index = self.find_tuple(idnum, self.queue)
		assert self.queue[index][2] == 'current', "Job looks like it wasn't running."
		self.queue[index][2] = 'completed'

	def find_tuple(self, val1, q):
		for index,i in enumerate(q):
			x, y, z = i
			if x == val1:
				return index



""" Worker (module)

This module contains a class that is designed to be spawned in
numbers as part of a queue of jobs to be completed. Each worker
carries out only one job, then dies.

The most common implementation of this module is to import it and
use inheritance to define how the worker should "work". The command
object it is passed is arbitrary, so the programmer is free to
define how each worker should interact with the object it is given.

The module has been structured to be used while threading, and
as a result contains a significant amount of threading logic
that remains application-agnostic, even as the definition of
"work" is altered for different workers.
"""

# imports for logging
import logging

# import pash from local library
import pash

# configures debug logging
logging.basicConfig(filename='/tmp/worker.log', level=logging.DEBUG,
    format='%(asctime)s (%(threadName)-2s) %(message)s',)


class Worker(object):
    """ Worker (class)

    A class object meant to be overriden to specify how work
    should be carried out.

    Implements:
        - 'logging' from Python standard library
        - Resource locking: The worker is passed an object which
           is expected have the attributes of a lock and a list.
        - Result passing: The worker adds an object to the lock
           before releasing it. This object represents the result
           of his work.

    Attributes:
        command (object): Some object which is to be worked on or
            otherwise processed.
        lock ('locked_queue'): An object that is expected to have
            the properties of a lock from the 'threading' module,
            as well as being capable of being appended to.

    """
    # setting the worker lock and the command to be carried out
    def __init__(self, lock, command):
        self.lock = lock
        self.command = command

    # container for the work being done, handles threading implementation, etc.
    def run(self):
        logging.debug('Starting a worker thread...')

        # acquire the lock
        self.lock.acquire()
        logging.debug('Got the lock.')

        # do work
        try:
            logging.debug("I am doing a thing: %s", self.command)
        except TypeError:
            logging.debug("I am doing something (that is not a string)")
        value = self.work(self.command)
        try:
            logging.debug("I did my thing: %s", self.command)
        except TypeError:
            logging.debug("I did something (that was not a string)")

        # add results to the worker lock, queen should pick these up
        self.lock.append(value)

        # release the lock
        logging.debug("Releasing the lock...")
        self.lock.release()

    # override this to define how work should be done;
    # this default implementation runs a bash command
    # and returns stdout
    def work(self, comm):
        """ Working

        This function should be overriden by the caller in order
        to provide the functionality of interest. The default
        implementation here uses 'pash' to run a command using
        Python's 'subprocess' module.

        This function is called during 'run', which is the intended
        target of any threading process. This function should
        never need to be called explicitly.

        Arguments:
            comm (string): A re-representation of 'command' from
                the 'run' function in this class

        Returns:
            proc.get_val('stdout') (string): The value of stdout
                returned for the command that was run
        """
        proc = pash.ShellProc()
        proc.run(comm)
        try:
       	    return proc.get_val('stdout').split('/n')
        except AttributeError:
            return None


